'''Main UI application'''

import io
import re
import os
import gramex
import gramex.cache
import gramex.handlers

# B404:import_subprocess only for JS compilation
import subprocess  # noqa S404
from hashlib import md5
from tornado.gen import coroutine, Return
from functools import partial
from gramex.config import variables, app_log, merge
from urllib.parse import urlparse, parse_qs, urlencode


def _join(*args):
    return os.path.normpath(os.path.join(*args))


ui_dir = os.path.dirname(os.path.abspath(__file__))
config_file = _join(ui_dir, 'config.yaml')
cache_dir = _join(variables['GRAMEXDATA'], 'apps', 'ui')
sass_bin = _join(ui_dir, 'node_modules', 'sass', 'sass.js')
if not os.path.exists(cache_dir):
    os.makedirs(cache_dir)

esbuild_path = _join(ui_dir, 'node_modules', 'esbuild', 'bin', 'esbuild')
# if esbuild is a shell script, run the base script
esbuild_cmd = [esbuild_path]
try:
    with open(esbuild_path, 'rb') as handle:
        if handle.read(2) == b'#!':
            esbuild_cmd = ['node', esbuild_path]
except OSError:
    pass


def cdn_redirect(handler, folder_map={}):
    '''Redirect /ui/... to cdn.jsdelivr.net/npm/...

    Before Gramex 1.84, specific npm libraries were included in Gramex. Now, we encourage projects
    to specify their own npm dependencies. This redirection makes apps backward compatible.

    By default, the path `/d3/dist/d3.min.js` is redirected to
    <https://cdn.jsdelivr.net/npm/d3/dist/d3.min.js>. But `folder_map` (specified in
    `gramex/apps/ui/gramex.yaml`) provides a mapping of folder names used earlier in Gramex to
    specific libraries, e.g. `{'d3v5': 'd3@5'}`.

    It redirects temporarily (HTTP 302). Add `?v=...` to URL for permanent redirection (HTTP 301).
    '''
    path = handler.path_args[0]
    for prefix, sub in folder_map.items():
        if path == prefix or path.startswith(prefix + '/'):
            path = sub + path[len(prefix) :]
            break
    handler.redirect(f'https://cdn.jsdelivr.net/npm/{path}', permanent=handler.get_arg('v', None))


def _get_cache_key(state):
    '''Return short string capturing state of object. Used to create unique filenames for state'''
    cache_key = gramex.cache.cache_key(state).encode('utf-8')
    # B303, B324:md5 is safe here - it's not for cryptographic use
    return md5(cache_key).hexdigest()[:5]  # noqa S113


@coroutine
def sass(
    handler: gramex.handlers.FileHandler, template: str = _join(ui_dir, 'bootstrap-theme.scss')
):
    '''Return a bootstrap theme based on the custom SASS variables provided.'''
    args = dict(variables.get('ui-bootstrap', {}))
    args.update({key: handler.get_arg(key) for key in handler.args})
    args = {key: val for key, val in args.items() if val}

    # Set default args
    config = gramex.cache.open(config_file)
    merge(args, config.get('defaults'), mode='setdefault')

    cache_key = _get_cache_key({'template': template, 'args': args})

    # Replace fonts from config file, if available
    google_fonts = set()
    for key in ('font-family-base', 'headings-font-family'):
        if key in args and args[key] in config['fonts']:
            fontinfo = config['fonts'][args[key]]
            args[key] = fontinfo['stack']
            if 'google' in fontinfo:
                google_fonts.add(fontinfo['google'])

    # Cache based on the dict and config as template.<cache-key>.css
    base = os.path.splitext(os.path.basename(template))[0] + '.' + cache_key
    cache_file = _join(cache_dir, base + '.css')
    if not os.path.exists(cache_file) or os.stat(template).st_mtime > os.stat(cache_file).st_mtime:
        # Create a SCSS file based on the args
        scss_path = _join(cache_dir, base + '.scss')
        with io.open(scss_path, 'wb') as handle:
            result = gramex.cache.open(template, 'template').generate(
                variables=args,
                google_fonts=google_fonts,
            )
            handle.write(result)
        # Run sass to generate the output
        proc = gramex.cache.Subprocess(
            [
                'node',
                sass_bin,
                scss_path,
                cache_file,
                '--style',
                'compressed',
                # Allow importing path from these paths
                '--load-path',
                os.path.dirname(template),
                '--load-path',
                ui_dir,
                '--load-path',
                bootstrap_dir,
            ]
        )
        out, err = yield proc.wait_for_exit()
        if proc.proc.returncode:
            raise RuntimeError('sass compilation failure', err.decode('utf-8'))

    handler.set_header('Content-Type', 'text/css')
    raise Return(gramex.cache.open(cache_file, 'bin', mode='rb'))


bootstrap_dir = _join(ui_dir, 'node_modules', 'bootstrap', 'scss')
# We only allow alphanumeric SASS keys (though SASS allows more)
valid_sass_key = re.compile(r'[_a-zA-Z][_a-zA-Z0-9\-]*')


@coroutine
def sass2(
    handler: gramex.handlers.FileHandler, path: str = _join(ui_dir, 'gramexui.scss')
) -> bytes:
    '''Compile a SASS file using custom variables from URL query parameters.

    Examples:
        >>> sass2(handler, 'x.sass')

    Parameters:
        handler: the[FileHandler][gramex.handlers.FileHandler] serving this file
        path: absolute path of input SASS file to compile into CSS

    Returns:
        compiled CSS file or source map if ?_map is specified

    URL query parameters in `handler.args` are converted into SASS variables.
    For example, `?primary=red` becomes `primary: red;` at the start of the SASS file.

    You can specify
    [`?@import=`](https://sass-lang.com/documentation/at-rules/import),
    [`?@use=`](https://sass-lang.com/documentation/at-rules/use) and
    [`?@forward=`](https://sass-lang.com/documentation/at-rules/forward)
    with 1 or more URLs. These URLs are imported as libraries.
    '''
    # Get valid variables from URL query parameters
    vars, commands, theme_colors = {}, {}, []
    for key, vals in handler.args.items():
        if key in {'@import', '@use', '@forward'}:
            commands[key] = vals
        # Ignore ?_map which is used for sourceMappingURL
        elif key == '_map':
            pass
        # Allow only alphanumeric SASS keys
        elif not valid_sass_key.match(key):
            app_log.warning('sass: "${key}" key not allowed. Use alphanumeric')
        # Pick the last arg, if it's non-empty
        elif len(vals) and vals[-1]:
            vars[key] = vals[-1]
            # ?color-alpha=red creates theme colors like .bg-alpha, .text-alpha, etc.
            # color_ is also supported for Python keyword arguments
            if key.startswith('color-') or key.startswith('color_'):
                theme_colors.append(f'"{key[6:]}": ${key}')
    if theme_colors:
        vars['theme-colors'] = f'({", ".join(theme_colors)})'

    # Create cache key based on state = path + imports + args. Output to <cache-key>.css
    path = os.path.normpath(path).replace('\\', '/')
    cache_key = _get_cache_key([path, commands, vars])
    target = _join(cache_dir, f'theme-{cache_key}.css')
    source = target[:-4] + '.scss'

    # Recompile if target is missing, or path has been updated
    if not os.path.exists(target) or os.stat(path).st_mtime > os.stat(target).st_mtime:
        # Create an SCSS file
        # ... whose contents include all variables
        content = [f'${key}: {val};' for key, val in vars.items()]
        # ... and commands @import, @use, @forward (convert \ to / to handle Windows paths)
        content += [
            '%s "%s";' % (key, url.replace('\\', '/'))
            for key, urls in commands.items()
            for url in urls
        ]
        # ... and the main SCSS file we want to use
        content.append(f'@import "{path}";')
        with open(source, 'w', encoding='utf-8') as handle:
            handle.write('\n'.join(content))
        # Compile SASS file. Allow @import from template dir, UI dir, Bootstrap dir, CWD
        proc = yield gramex.service.threadpool.submit(
            subprocess.run,
            [
                'node',
                sass_bin,
                source,
                target,
                '--style',
                'compressed',
                '--load-path',
                os.path.dirname(path),
                '--load-path',
                ui_dir,
                '--load-path',
                bootstrap_dir,
                '--load-path',
                os.getcwd(),
            ],
            capture_output=True,
            input='\n'.join(content),
            encoding='utf-8',
        )
        if proc.returncode:
            # If there's an error, remove the generated files and raise an error
            os.remove(source)
            raise RuntimeError(f'.sass compilation failure:\n{proc.stderr}\n{proc.stdout}')
    return _sourcemap(handler, target, 'text/css')


@coroutine
def jscompiler(
    handler: gramex.handlers.FileHandler,
    path: str,
    target_ext: str,
    cmd: str,
    options: dict = {},
) -> bytes:
    '''Compile a file (e.g. TypeScript) into a JS file using Node.js

    Examples:
        >>> jscompiler(
        ...     handler, path='x.ts', target_ext='.js', exe='/path/to/esbuild',
        ...     cmd=lambda filename, target_dir, options: ...)

    Parameters:
        handler: the[FileHandler][gramex.handlers.FileHandler] serving this file
        path: absolute path of input file to compile into JavaScript
        target_ext: extension of output file (e.g. `.js`, `.min.js`)
        cmd: command line to run. This substitutes 3 variables:
            - `$exe` for the `exe` parameter
            - `$filename` for the absolute path to the input file
            - `$targetDir` for the absolute path to the output directory

    Returns:
        compiled JS file or source map if ?_map is specified
    '''
    # Get valid variables from URL query parameters
    # Create cache key based on state = path. Output to <cache-key>.js
    path = os.path.normpath(path).replace('\\', '/')
    ext = os.path.splitext(path)[-1]
    options = [f'--{key}={handler.get_arg(key, val)}' for key, val in options.items()]
    cache_key = _get_cache_key([path] + options)
    target_dir = _join(cache_dir, f'{ext}-{cache_key}')
    target = os.path.join(target_dir, os.path.basename(path[: -len(ext)] + target_ext))

    # Recompile if output target is missing, or path has been updated
    if not os.path.exists(target) or os.stat(path).st_mtime > os.stat(target).st_mtime:
        cwd, filename = os.path.split(path)
        cmd_to_run = cmd(filename=filename, target_dir=target_dir, options=options)
        app_log.debug(f'Compiling {ext}: {cmd_to_run}')
        proc = yield gramex.service.threadpool.submit(
            subprocess.run,
            cmd_to_run,
            cwd=cwd,
            capture_output=True,
            encoding='utf-8',
        )
        if proc.returncode:
            raise RuntimeError(f'.{ext} compilation failure:\n{proc.stderr}\n{proc.stdout}')

    return _sourcemap(handler, target, 'text/javascript')


ts = partial(
    jscompiler,
    target_ext='.js',
    options={
        'format': 'iife',
        'bundle': 'true',
        'minify': 'true',
        'target': 'esnext',
        'charset': 'utf8',
        'global-name': '',
        'keep-names': 'false',
    },
    cmd=lambda options, filename, target_dir: [
        *esbuild_cmd,
        *options,
        '--allow-overwrite',
        '--sourcemap',
        f'--outdir={target_dir}',
        filename,
    ],
)


def _sourcemap(handler: gramex.handlers.FileHandler, target: str, mime: str) -> bytes:
    '''Returns the compiled target file OR the source map if ?_map is set.

    Examples:
        >>> _sourcemap(handler, 'output.js', 'text/javascript')

    Parameters:
        handler: the [FileHandler][gramex.handlers.FileHandler] serving this file
        target: absolute path of compiled output
        mime: MIME type of compiled output

    Returns:
        source map or target file contents

    This is used by FileHandlers compiling TS, SASS, etc.

    If the URL has a ?_map, it serves `{target}.map` as a JSON file.
    Else it serves the `{target}` as `mime` type,
    replacing `sourceMappingURL` with the current URL + `?_map`.
    '''
    if '_map' in handler.args:
        # Serve JSON source map if requested
        handler.set_header('Content-Type', 'application/json')
        return gramex.cache.open(target + '.map', 'bin', mode='rb')
    else:
        # Serve compiled file if there's no ?_map
        handler.set_header('Content-Type', mime)
        content = gramex.cache.open(target, 'bin', mode='rb')
        # ... but replace the sourceMappingURL with URL + ?_map
        url = urlparse(handler.request.uri)
        query = parse_qs(url.query, keep_blank_values=True)
        query.setdefault('_map', [''])
        source_map = os.path.basename(url.path) + '?' + urlencode(query, doseq=True)
        return re.sub(
            rb' sourceMappingURL=(\S+)',
            rb' sourceMappingURL=' + source_map.encode('utf-8'),
            content,
        )



import os
import tornado.gen
import tornado.template
import gramex.cache


@tornado.gen.coroutine
def template(content, handler=None, **kwargs):
    '''Renders a file as a Tornado template in FileHandler.

    This can be used as a keyword argument to FileHandler:

    ```yaml
    pattern: /file.tmpl.html
    handler: FileHandler
    kwargs:
        ...
        template: '*.tmpl.html'
    ```
    '''
    name = f'{handler.name}:template'
    loader = None
    # If a FileHandler renders templates, cache the file, and treat handler.file as filename.
    if handler is not None and getattr(handler, 'file', None):
        name = str(handler.file)
        loader = CacheLoader(os.path.dirname(str(handler.file)))
    tmpl = tornado.template.Template(content, name=name, loader=loader)

    if handler is not None:
        for key, val in handler.get_template_namespace().items():
            kwargs.setdefault(key, val)

    # handler is added to kwargs by handler.get_template_namespace()
    return tmpl.generate(**kwargs).decode('utf-8')


@tornado.gen.coroutine
def scss(content, handler):
    '''Renders a SCSS file as CSS in a FileHandler.

    This can be used as a keyword argument to FileHandler:

    ```yaml
    pattern: /file.scss
    handler: FileHandler
    kwargs:
        ...
        scss: '*.scss'
    ```
    '''
    from gramex.apps.ui import sass2

    # Ignore the content provided. sass2 needs the file actually located at handler.path.
    result = yield sass2(handler, handler.path)
    return result.decode('utf-8')


# SCSS files and SASS files are compiled exactly the same way
sass = scss


@tornado.gen.coroutine
def ts(content, handler):
    '''Render a TypeScript file as JavaScript in a FileHandler.

    This can be used as a keyword argument to FileHandler:

    ```yaml
    pattern: /file.ts
    handler: FileHandler
    kwargs:
        ...
        ts: '*.ts'
    ```
    '''
    from gramex.apps.ui import ts

    # Ignore the content provided. ts needs the file actually located at handler.path.
    result = yield ts(handler, handler.path)
    return result.decode('utf-8')


class CacheLoader(tornado.template.Loader):
    # Like tornado.template.Loader, but caching only until underlying file is changed.
    # Used internally by BaseHandler to override the Tornado default template loader.

    def load(self, name, parent_path=None):
        # Always load the file. _create_template takes care of the caching
        name = self.resolve_path(name, parent_path=parent_path)
        with self.lock:
            self.templates[name] = self._create_template(name)
            return self.templates[name]

    def _template_opener(self, path):
        with open(path, 'rb') as f:
            return tornado.template.Template(f.read(), name=self._name, loader=self)

    def _create_template(self, name):
        # Use gramex.cache.open to ensure that the file is cached
        self._name = name
        return gramex.cache.open(os.path.join(self.root, name), self._template_opener)



import re
import os
import time
import shlex
import atexit
import psutil
import requests
import tornado.gen
from orderedattrdict import AttrDict
from threading import Thread, Lock

# B404:import_subprocess only for JS compilation
from subprocess import Popen, PIPE, STDOUT  # noqa S404
from urllib.parse import urlencode, urljoin
from tornado.web import HTTPError
from tornado.httpclient import AsyncHTTPClient
from gramex.config import app_log, variables
from gramex.cache import cache_key
from gramex.http import OK, BAD_REQUEST, GATEWAY_TIMEOUT, BAD_GATEWAY, CLIENT_TIMEOUT
from .basehandler import BaseHandler

_PPTX_MIME = 'application/vnd.openxmlformats-officedocument.presentationml.presentation'
# HTTP headers not to forward to chromecapture.js.
# Keep this sync-ed with the same list in chromecapture.js
_IGNORE_HEADERS = {
    'host',  # The URL will determine the host
    'connection',  # Let Tornado manage the connection
    'upgrade',  # .. and the upgrades
    'content-length',  # The new request will have a different content - length
    'content-md5',  # ... and different content - md5
}
DEFAULT_TIMEOUT = 10


class Capture:
    default_port = 9900  # Default port to run CaptureJS at
    check_interval = 0.05  # Frequency (seconds) to check if self.started
    # Set engine configurations for PhantomJS and Puppeteer
    engines = AttrDict(
        phantomjs=AttrDict(
            cmd='phantomjs --ssl-protocol=any',
            script='capture.js',
            first_line=b'PhantomJS.*capture\\.js',
            name='Capture',
            version='1.0',
        ),
        chrome=AttrDict(
            cmd='node',
            script='chromecapture.js',
            first_line=b'node\\.js.*chromecapture\\.js',
            name='ChromeCapture',
            version='1.1',
        ),
    )

    def __init__(
        self,
        port: int = None,
        url: str = None,
        engine: str = None,
        cmd: str = None,
        timeout: int = DEFAULT_TIMEOUT,
    ):
        '''
        Create a proxy for capture.js.

        Examples:
            >>> capture = Capture()
            >>> with open('screenshot.png', 'wb') as handle:
            ...    handle.write(capture.png('https://gramener.com/'))
            >>> with open('screenshot.pdf', 'wb') as handle:
            ...    handle.write(capture.pdf('https://gramener.com/'))

        Parameters:
            port: port where capture.js is running. Default: 9900
            url: URL:port where PhantomJS is running with capture.js.
                Default: `http://localhost:<port>/`
            cmd: Command to run PhantomJS with capture.js at the specified
                port. Default: `phantomjs $GRAMEXPATH/apps/capture/capture.js --port=<port>`
            timeout: Seconds to wait for PhantomJS to timeout. Default: 10

        The constructor runs `Capture.start()` in a new thread, which checks if
        capture.js is running at `url`. If not, it runs `cmd` and checks again.
        Until capture.js is detected, all capture methods will fail.
        '''
        # Set default values for port, url and cmd
        self.engine = self.engines['phantomjs' if engine is None else engine]
        port = self.default_port if port is None else port
        if url is None:
            url = f'http://localhost:{port}/'
            if cmd is None:
                script = os.path.join(variables.GRAMEXPATH, 'apps', 'capture', self.engine.script)
                cmd = f'{self.engine.cmd} "{script}" --port={port}'
        self.url = url
        self.first_line_re = re.compile(self.engine.first_line)
        self.cmd = cmd
        self.timeout = timeout
        self.browser = AsyncHTTPClient()
        self.lock = Lock()
        self.started = False
        self.start()

    def start(self):
        '''
        Starts a thread and check if capture is already running at `url`. If
        not, start `cmd` and check again. Print logs from `cmd`.

        This method is thread-safe. It may be called as often as required.
        CaptureHandler calls this method if `?start` is passed.
        '''
        with self.lock:
            thread = Thread(target=self._start, name=f'Capture {self.engine} @ {self.url}')
            thread.daemon = True
            thread.start()

    def _start(self):
        '''
        Check if capture is already running at `url`. If not, start `cmd`
        and check again. Print logs from `cmd`.
        '''
        self.started = False
        script = self.engine.script
        try:
            # Check if capture.js is at the url specified
            app_log.info(f'Pinging {script} at {self.url}')
            r = requests.get(self.url, timeout=self.timeout)
            self._validate_server(r)
            self.started = True
        except requests.ReadTimeout:
            # If capture.js doesn't respond immediately, we haven't started
            app_log.error(f'url: {self.url} timed out')
        except requests.ConnectionError:
            # Try starting the process again
            app_log.info(f'Starting {script} via {self.cmd}')
            self.close()
            # B603:subprocess_without_shell_equals_true is safe since self.cmd is taken from
            # the YAML configuration (from developers)
            self.proc = Popen(shlex.split(self.cmd), stdout=PIPE, stderr=STDOUT)  # noqa S603
            self.proc.poll()
            atexit.register(self.close)
            # TODO: what if readline() does not return quickly?
            line = self.proc.stdout.readline().strip()
            if not self.first_line_re.search(line):
                return app_log.error(f'cmd: {self.cmd} invalid. Returned "{line}"')
            app_log.info(f'Pinging {script} at {self.url}')
            try:
                r = requests.get(self.url, timeout=self.timeout)
                self._validate_server(r)
                pid = self.proc.pid
                app_log.info(line.decode('utf-8') + f' live (pid={pid})')
                self.started = True
                # Keep logging capture.js output until proc is killed by another thread
                while hasattr(self, 'proc'):
                    line = self.proc.stdout.readline().strip()
                    if len(line) == 0:
                        app_log.info(f'{script} terminated: pid={pid}')
                        self.started = False
                        break
                    # Capture won't print anything, unless there's a problem, or if debug is on.
                    # So log it at warning level not info.
                    app_log.warning(line.decode('utf-8'))
            except Exception:
                app_log.exception(f'Ran {self.cmd}. But {script} not at {self.url}')
        except Exception:
            app_log.exception('Cannot start Capture')

    def close(self):
        '''Stop capture.js if it has been started by this object'''
        if hasattr(self, 'proc'):
            try:
                process = psutil.Process(self.proc.pid)
                for proc in process.children(recursive=True):
                    proc.kill()
                process.kill()
            except psutil.NoSuchProcess:
                app_log.info(f'{self.engine.script} PID {self.proc.pid} already killed')
                pass
            delattr(self, 'proc')

    def _validate_server(self, response):
        # Make sure that the response we got is from the right version of capture.js
        server = response.headers.get('Server', '')
        parts = server.split('/', 2)
        script = self.engine.script
        if len(parts) != 2 or parts[0] != self.engine.name or parts[1] < self.engine.version:
            raise RuntimeError(f'Server: {server} at {self.url} is not {script}')

    @tornado.gen.coroutine
    def capture_async(self, headers=None, **kwargs):
        '''
        Returns a screenshot of the URL. Runs asynchronously in Gramex. Arguments
        are same as [capture][gramex.handlers.capturehandler.Capture.capture]
        '''
        # If ?start is provided, start server and wait until timeout
        if 'start' in kwargs:
            self.start()
            end_time = time.time() + self.timeout
            while not self.started and time.time() < end_time:
                yield tornado.gen.sleep(self.check_interval)
        if not self.started:
            raise RuntimeError(f'{self.engine.script} not started. See logs')
        r = yield self.browser.fetch(
            self.url,
            method='POST',
            body=urlencode(kwargs, doseq=True),
            raise_error=False,
            connect_timeout=self.timeout,
            request_timeout=self.timeout,
            headers=headers,
        )
        if r.code == OK:
            self._validate_server(r)
        raise tornado.gen.Return(r)

    def capture(self, url, **kwargs):
        '''
        Return a screenshot of the URL.

        Parameters:
            url (str): URL to take a screenshot of
            ext (str): format of output. Can be pdf, png, gif or jpg
            selector (str): Restrict screenshot to (optional) CSS selector in URL
            delay (int): milliseconds (or expression) to wait for before taking a screenshot
            format (str): A3, A4, A5, Legal, Letter or Tabloid. Defaults to A4. For PDF
            layout (str): A3, A4, A5, Legal, 16x9, 16x10, 4x3. Defaults to 4x3. For PPTX
            orientation (str): portrait or landscape. Defaults to portrait. For PDF
            header (str): header for the page. For PDF
            footer (str): footer for the page. For PDF
            width (int): screen width. Default: 1200. For PNG/GIF/JPG
            height (int): screen height. Default: 768. For PNG/GIF/JPG
            scale (float): zooms the screen by a factor. For PNG/GIF/JPG
            dpi (int): dots (pixels) per inch. For PPTX
            title (str): slide title. For PPTX
            debug (int): sets log level for HTTP requests (2) and responses (1)

        Returns:
            a bytestring with the binary contents of the screenshot

        Raises:
            RuntimeError: if capture.js is not running or fails
        '''
        # Ensure that we're connecting to the right version of capture.js
        if not self.started:
            end_time = time.time() + self.timeout
            while not self.started and time.time() < end_time:
                time.sleep(self.check_interval)
            if not self.started:
                raise RuntimeError(f'{self.engine.script} not started. See logs')
        kwargs['url'] = url
        r = requests.post(self.url, data=kwargs, timeout=self.timeout)
        if r.status_code == OK:
            self._validate_server(r)
            return r.content
        else:
            raise RuntimeError(f'{self.engine.script} error: {r.content}')

    def pdf(self, url, **kwargs):
        '''An alias for `Capture.capture()` with `ext='pdf'`.'''
        kwargs['ext'] = 'pdf'
        return self.capture(url, **kwargs)

    def png(self, url, **kwargs):
        '''An alias for `Capture.capture()` with `ext='png'`.'''
        kwargs['ext'] = 'png'
        return self.capture(url, **kwargs)

    def pptx(self, url, **kwargs):
        '''An alias for `Capture.capture()` with `ext='pptx'`.'''
        kwargs['ext'] = 'pptx'
        return self.capture(url, **kwargs)

    def jpg(self, url, **kwargs):
        '''An alias for `Capture.capture()` with `ext='jpg'`.'''
        kwargs['ext'] = 'jpg'
        return self.capture(url, **kwargs)

    def gif(self, url, **kwargs):
        '''An alias for `Capture.capture()` with `ext='gif'`.'''
        kwargs['ext'] = 'gif'
        return self.capture(url, **kwargs)


class CaptureHandler(BaseHandler):
    '''
    Renders a web page as a PDF or as an image. It accepts the same arguments as
    `Capture`.

    The page is called with the same args as `Capture.capture()`. It also
    accepts a `?start` parameter that restarts capture.js if required.
    '''

    # Each config maps to a Capture() object. cls.captures[config] = Capture()
    captures = {}

    @classmethod
    def setup(
        cls,
        port=None,
        url=None,
        engine=None,
        cmd=None,
        timeout=DEFAULT_TIMEOUT,
        pattern=None,
        **kwargs,
    ):
        super(CaptureHandler, cls).setup(**kwargs)
        # Create a new Capture only if the config has changed.
        config = {'engine': engine, 'port': port, 'url': url, 'cmd': cmd, 'timeout': timeout}
        config_str = cache_key(config)
        if config_str not in cls.captures:
            cls.captures[config_str] = Capture(**config)
        cls.capture = cls.captures[config_str]
        # TODO: if the old config is no longer used, close it
        cls.ext = {
            'pdf': {'mime': 'application/pdf'},
            'png': {'mime': 'image/png'},
            'jpg': {'mime': 'image/jpeg'},
            'jpeg': {'mime': 'image/jpeg'},
            'gif': {'mime': 'image/gif'},
            'pptx': {'mime': _PPTX_MIME},
        }
        cls.pattern = pattern

    @tornado.gen.coroutine
    def get(self):
        args = self.argparse(
            url={'default': self.request.headers.get('Referer', None)},
            ext={'choices': self.ext, 'default': 'pdf'},
            file={'default': 'screenshot'},
            emulate={},
            selector={'nargs': '*'},
            cookie={},
            delay={},
            width={'type': int},
            height={'type': int},
            x={'type': int},
            y={'type': int},
            scale={'type': float},
            dpi={'type': int, 'nargs': '*'},
            format={'choices': ['A3', 'A4', 'A5', 'Legal', 'Letter', 'Tabloid'], 'default': 'A4'},
            layout={'choices': ['A3', 'A4', 'Letter', '16x9', '16x10', '4x3'], 'default': '4x3'},
            orientation={'choices': ['portrait', 'landscape'], 'default': 'portrait'},
            title={'nargs': '*'},
            title_size={'type': int, 'nargs': '*'},
            start={'nargs': '*'},
            debug={'nargs': '*'},
            header={},
            footer={},
            headerTemplate={},
            footerTemplate={},
            margins={},
        )
        if args['url'] is None:
            raise HTTPError(BAD_REQUEST, f'{self.name}: CaptureHandler needs ?url=')

        # ?url= can be a relative URL. Use the full X-Request URL as the base
        args['url'] = urljoin(self.xrequest_full_url, args['url'])

        # Pass configured timeout to capture.js to await PDF generation
        args['timeout'] = self.capture.timeout

        if self.pattern and not re.search(self.pattern, args['url']):
            error = f'{self.name}: {args["url"]} does not match pattern {self.pattern}'
            raise HTTPError(BAD_REQUEST, error)

        # Copy all relevant HTTP headers as-is
        args['headers'] = {
            key: val for key, val in self.request.headers.items() if key not in _IGNORE_HEADERS
        }
        if 'cookie' not in args:
            cookie = self.request.headers.get('Cookie', None)
            if cookie is not None:
                args['cookie'] = cookie
        info = self.ext[args.ext]
        try:
            response = yield self.capture.capture_async(**args)
        except RuntimeError as e:
            # capture.js could not fetch the response
            raise HTTPError(BAD_GATEWAY, e.args[0])

        if response.code == OK:
            self.set_header('Content-Type', info['mime'])
            self.set_header(
                'Content-Disposition', 'attachment; filename="{file}.{ext}"'.format(**args)
            )
            self.write(response.body)
        elif response.code == CLIENT_TIMEOUT:
            self.set_status(GATEWAY_TIMEOUT, 'Capture is busy')
            self.set_header('Content-Type', 'application/json')
            self.write(
                {
                    'status': 'fail',
                    'msg': [f'Capture did not respond within timeout: {self.capture.timeout}s'],
                }
            )
        else:
            self.set_status(response.code, 'capture.js error')
            self.set_header('Content-Type', 'application/json')
            self.write(response.body)



import json
import os
import requests
import pandas as pd
import gramex
import gramex.handlers
from . import server
from . import TestGramex
from orderedattrdict import AttrDict
from nose.tools import eq_, ok_, assert_raises
from tornado.web import create_signed_value
from urllib.request import urlopen
from urllib.error import HTTPError
from gramex.services import info
from gramex.http import OK, NOT_FOUND, INTERNAL_SERVER_ERROR, FORBIDDEN, TOO_MANY_REQUESTS


class TestURLPriority(TestGramex):
    # Test Gramex URL priority sequence

    def test_url_priority(self):
        self.check('/path/abc', text='/path/.*')
        self.check('/path/file', text='/path/file')
        self.check('/path/dir', text='/path/.*')
        self.check('/path/dir/', text='/path/dir/.*')
        self.check('/path/dir/abc', text='/path/dir/.*')
        self.check('/path/dir/file', text='/path/dir/file')
        self.check('/path/priority', text='/path/priority')


class TestURLNormalization(TestGramex):
    # Test URL pattern normalization

    def test_url_normalization(self):
        self.check('/path/norm1', text='/path/norm1')
        self.check('/path/norm2', text='/path/norm2')


class TestAttributes(TestGramex):
    # Ensure that BaseHandler subclasses have relevant attributes

    def test_attributes(self):
        self.check('/func/attributes', code=OK)


class TestXSRF(TestGramex):
    # Test BaseHandler xsrf: setting

    def test_xsrf(self):
        r = self.check('/path/norm')
        ok_('Set-Cookie' not in r.headers)

        # First request sets xsrf cookie
        session = requests.Session()
        r = session.get(server.base_url + '/xsrf', timeout=10)
        ok_('Set-Cookie' in r.headers)
        ok_('_xsrf' in r.headers['Set-Cookie'])

        # Next request does not set xsrf cookie, because it already exists
        r = session.get(server.base_url + '/xsrf', timeout=10)
        ok_('Set-Cookie' not in r.headers)

    def test_xsrf_false(self):
        # When xsrf_cookies is set to False, POST works
        r = requests.post(server.base_url + '/xsrf/no')
        eq_(r.status_code, OK)

    def test_xsrf_true(self):
        # When xsrf_cookies is set to True, POST fails without _xsrf
        r = requests.post(server.base_url + '/xsrf/yes')
        eq_(r.status_code, FORBIDDEN)

    def test_xsrf_ajax(self):
        # Requests sent with X-Requested-With should not need an XSRF cookie
        url = server.base_url + '/xsrf/yes'
        # Mangle case below to ensure Gramex handles it case-insensitively
        r = requests.post(url, headers={'X-Requested-With': 'xMlHtTpReQuESt'})
        eq_(r.status_code, OK)
        # fetch() sends a Sec-Fetch-Mode: cors
        r = requests.post(url, headers={'Sec-Fetch-Mode': 'cors'})
        eq_(r.status_code, OK)
        # <form> sends a Sec-Fetch-Mode of navigate in modern (desktop) browsers.
        # Older browsers do not send it.
        r = requests.post(url, headers={'Sec-Fetch-Mode': 'navigate'})
        eq_(r.status_code, FORBIDDEN)


class TestSetupErrors(TestGramex):
    def test_invalid_handler(self):
        self.check('/invalid-handler', code=NOT_FOUND)

    def test_invalid_setup(self):
        self.check(
            '/invalid-setup',
            code=INTERNAL_SERVER_ERROR,
            text='url:invalid-setup: needs "function:"',
        )

    def test_invalid_function(self):
        self.check('/invalid-function', code=INTERNAL_SERVER_ERROR, text='nonexistent')


class TestErrorHandling(TestGramex):
    # Test BaseHandler error: setting
    def test_404_escaped(self):
        # Check that templates are HTML escaped by default
        try:
            # Requests converts <script> into %3Cscript%3E before sending URL.
            # So use urlopen instead of requests.get
            urlopen(server.base_url + '/error/404-escaped-<script>')
        except HTTPError as err:
            eq_(err.code, NOT_FOUND)
            text = err.read().decode('utf-8')
            ok_(' &quot;/error/404-escaped-&lt;script&gt;&quot;' in text)
            ok_('\n' in text)  # error-404.json template has newlines
        else:
            ok_(False, '/error/404-escaped-<script> should raise a 404')

    def test_404_unescaped(self):
        # autoescape can be over-ridden
        try:
            urlopen(server.base_url + '/error/404-template-<script>')
        except HTTPError as err:
            eq_(err.code, NOT_FOUND)
            text = err.read().decode('utf-8')
            ok_(' "/error/404-template-<script>' in text)
            ok_('\n' not in text)  # since whitespace=oneline
        else:
            ok_(False, '/error/404-template-<script> should raise a 404')

    def test_500(self):
        r = self.check('/error/500-function', code=INTERNAL_SERVER_ERROR)
        eq_(r.headers['Content-Type'], 'application/json')
        result = r.json()
        eq_(result['status_code'], INTERNAL_SERVER_ERROR)
        ok_(result['handler.request.uri'].endswith('/error/500-function'))


class TestMime(TestGramex):
    def setUp(self):
        self.mime_map = {
            '.yml': 'text/yaml; charset=UTF-8',
            # ToDo: Fix with FileHandler 2
            # '.yaml': 'text/yaml; charset=UTF-8',
            '.md': 'text/markdown; charset=UTF-8',
            '.markdown': 'text/markdown; charset=UTF-8',
            '.json': 'application/json',
            '.svg': 'image/svg+xml',
            # '.py': 'text/plain; charset=UTF-8',       # .py files are forbidden by default
            '.h5': 'application/x-hdf5',
            '.hdf5': 'application/x-hdf5',
        }
        self.files = set()
        folder = os.path.dirname(os.path.abspath(__file__))
        for ext, mime in self.mime_map.items():
            path = os.path.join(folder, 'dir', 'gen' + ext)
            self.files.add(path)
            with open(path, 'wb'):
                pass

    def test_mime(self):
        for ext, mime in self.mime_map.items():
            r = self.check('/dir/gen' + ext)
            eq_(r.headers['Content-Type'], mime)

    def tearDown(self):
        for file in self.files:
            os.unlink(file)


class TestBaseHandler(TestGramex):
    def test_headers(self):
        self.check('/', headers={'X-BaseHandler': 'base'})

    def test_kwargs(self):
        # Fetch {name: handler_class} for Tornado 5.0 onwards
        classes = {
            rule.handler_class.name: rule.handler_class
            for rules in info.app.default_router.rules
            for rule in rules.target.rules
        }
        # FileHandler inherits kwargs from handlers.FileHandler and handlers.BaseHandler
        cls = classes['base']
        eq_(cls.name, 'base')
        eq_(cls.kwargs['path'], 'dir/index.html')
        headers = cls.kwargs['headers']
        eq_(headers['X-BaseHandler'], 'base')
        eq_(headers['X-BaseHandler'], 'base')
        eq_(headers['X-FileHandler'], 'base')
        eq_(headers['X-FileHandler-Base'], 'base')


class TestDefaultGramexYAML(TestGramex):
    def test_default_files(self):
        # Default gramex.yaml FileHandler exposes files in the current directory
        self.check('/README.md', path='README.md', headers={'Cache-Control': 'max-age=60'})
        self.check('/sample.png', path='sample.png', headers={'Cache-Control': 'max-age=60'})

    def test_favicon(self):
        self.check(
            '/favicon.ico',
            path='../gramex/favicon.ico',
            headers={'Cache-Control': 'public, max-age=86400'},
        )


class TestAlias(TestGramex):
    def test_alias(self):
        # gramex.service uses service: as an alias for handler.
        # It also creates alternate names for these services.
        # Just check a few. If some work, all should work.
        self.check('/alias-command')
        self.check('/alias-data')
        self.check('/alias-function')


class TestConfig(TestGramex):
    def test_appconfig(self):
        # gramex.appconfig should contain the current app configuration
        r = self.check('/appconfig').json()
        eq_(r['url']['appconfig']['pattern'], '/appconfig')
        eq_(r['url']['appconfig']['kwargs']['headers'], {'Content-Type': 'application/json'})


class TestRateLimit(TestGramex):
    def test_setup(self):
        def setup(ratelimit=None, ratelimit_app_conf=None, base=gramex.handlers.BaseHandler):
            '''Run setup_ratelimit() on a new subclass of cls'''
            class_vars = {
                'conf': AttrDict(pattern='/test'),
                'name': 'test',
                '_on_init_methods': [],
                '_on_finish_methods': [],
            }
            handler = type(base.__name__, (base,), class_vars)
            handler.setup_ratelimit(ratelimit, ratelimit_app_conf)
            return handler

        # ratelimit: is optional
        setup(None, None)

        # ratelimit: requires app.ratelimit
        with assert_raises(ValueError) as cm:
            setup({}, None)
        eq_(cm.exception.args[0], "url:test.ratelimit: no app.ratelimit defined")

        # ratelimit.keys required
        with assert_raises(ValueError) as cm:
            setup({}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], 'url:test.ratelimit.keys: missing')

        # ratelimit.limit required
        with assert_raises(ValueError) as cm:
            setup({'keys': 'daily'}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], 'url:test.ratelimit.limit: missing')

        # ratelimit.keys strings MUST be predefined
        with assert_raises(ValueError) as cm:
            setup({'keys': 'secondly', 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], 'url:test.ratelimit.keys: secondly is unknown')

        # ratelimit.keys strings MUST be str/list
        with assert_raises(ValueError) as cm:
            setup({'keys': 0, 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], 'url:test.ratelimit.keys: needs dict list, not 0')

        # ratelimit.keys: invalid function compilation directly raises exceptions, e.g. SyntaxError
        with assert_raises(SyntaxError):
            setup({'keys': [{'function': ';$@'}], 'limit': 5}, gramex.conf.app.ratelimit)

        # ratelimit.keys dicts need a function:
        with assert_raises(ValueError) as cm:
            setup({'keys': {'x': 0}, 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], "url:test.ratelimit.keys: {'x': 0} has no function/key:")
        with assert_raises(ValueError) as cm:
            setup({'keys': [{'x': 0}], 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], "url:test.ratelimit.keys: {'x': 0} has no function/key:")

        # ratelimit.keys can be a predefined key string
        for key in ('hourly', 'daily', 'weekly', 'monthly', 'yearly', 'user'):
            ok_(setup({'keys': key, 'limit': 5}, gramex.conf.app.ratelimit)._ratelimit[0].store)

        # ratelimit.keys can be a dict with a function. Defines a single key
        ok_(setup({'keys': {'function': '0'}, 'limit': 5}, gramex.conf.app.ratelimit))

        # ratelimit.keys can be a dict with a key
        ok_(setup({'keys': {'key': 'daily'}, 'limit': 5}, gramex.conf.app.ratelimit))

        # ratelimit.keys can be a comma-separated string
        cls = setup({'keys': 'daily, user', 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(len(cls._ratelimit[0].key_fn), 2)

        # ratelimit.keys can be an array of strings
        cls = setup({'keys': ['daily', 'user'], 'limit': 5}, gramex.conf.app.ratelimit)
        eq_(len(cls._ratelimit[0].key_fn), 2)

        # ratelimit.limit must be a number
        with assert_raises(ValueError) as cm:
            setup({'keys': 'daily', 'limit': 'x'}, gramex.conf.app.ratelimit)
        eq_(cm.exception.args[0], "url:test.ratelimit.limit: needs {'function': number}, not x")

        # ratelimit.limit can be a function
        cls = setup({'keys': 'user', 'limit': {'function': '5'}}, gramex.conf.app.ratelimit)
        eq_(cls._ratelimit[0].limit_fn(None), 5)

        # EVERY handler supports rate-limiting via `ratelimit:`
        for base in (
            gramex.handlers.FunctionHandler,
            gramex.handlers.FormHandler,
            gramex.handlers.FileHandler,
            gramex.handlers.DriveHandler,
        ):
            cls = setup({'keys': 'daily, user', 'limit': 5}, gramex.conf.app.ratelimit, base)
            ok_(issubclass(cls, base))

        # ratelimit.keys return the expected values
        all_keys = ['hourly', 'daily', 'weekly', 'monthly', 'yearly', 'user']
        cls = setup({'keys': all_keys, 'limit': 5}, gramex.conf.app.ratelimit)
        key_fn = dict(zip(all_keys, cls._ratelimit[0].key_fn))
        handler = AttrDict(current_user=AttrDict(id='a@b'))
        eq_(key_fn['hourly']['function'](handler), pd.Timestamp.utcnow().strftime('%Y-%m-%d %H'))
        eq_(key_fn['daily']['function'](handler), pd.Timestamp.utcnow().strftime('%Y-%m-%d'))
        eq_(key_fn['weekly']['function'](handler), pd.Timestamp.utcnow().strftime('%Y %U'))
        eq_(key_fn['monthly']['function'](handler), pd.Timestamp.utcnow().strftime('%Y-%m'))
        eq_(key_fn['yearly']['function'](handler), pd.Timestamp.utcnow().strftime('%Y'))
        eq_(key_fn['user']['function'](handler), 'a@b')

        # Multiple pools are supported
        cls = setup(
            [{'keys': 'daily, user', 'limit': 5}, {'keys': 'daily', 'limit': 10}],
            gramex.conf.app.ratelimit,
        )
        key_fn0, key_fn1 = cls._ratelimit[0].key_fn, cls._ratelimit[1].key_fn
        eq_(key_fn0[0]['function'](handler), pd.Timestamp.utcnow().strftime('%Y-%m-%d'))
        eq_(key_fn0[1]['function'](handler), 'a@b')
        eq_(key_fn1[0]['function'](handler), pd.Timestamp.utcnow().strftime('%Y-%m-%d'))
        eq_(len(key_fn1), 1)

        # If any pool has an error, entire class is not set up
        with assert_raises(ValueError) as cm:
            setup([{'keys': 'daily', 'limit': 10}, {'keys': 'weekly'}], gramex.conf.app.ratelimit)

    def check_rate(self, url, user, limit, remaining, code=OK, expiry='daily'):
        # If user= is specified, send an {'id': user} object via headers
        headers = {}
        if user is not None:
            cookie_secret = gramex.service.app.settings['cookie_secret']
            sign = create_signed_value(cookie_secret, 'user', json.dumps({'id': user}))
            headers['X-Gramex-User'] = sign

        # Check the status code and X-Ratelimit-* headers
        r = self.check(url, code=code, request_headers=headers)
        if code in (OK, TOO_MANY_REQUESTS):
            eq_(r.headers['X-Ratelimit-Limit'], str(limit))
            eq_(r.headers['X-Ratelimit-Remaining'], str(remaining))
            # Check expiry time to plus/minus 2 seconds.
            # NOTE: This test may fail if running exactly at UTC midnight
            if expiry in gramex.handlers.basehandler._PREDEFINED_KEYS:
                expiry = gramex.handlers.basehandler._PREDEFINED_KEYS[expiry]['expiry']()
            ok_(expiry - 2 <= int(r.headers['X-Ratelimit-Reset']) <= expiry + 2)
            # Retry-After header should be sent only for
            if code == TOO_MANY_REQUESTS:
                eq_(r.headers['Retry-After'], r.headers['X-Ratelimit-Reset'])
            else:
                ok_('Retry-After' not in r.headers)
        else:
            # No headers sent for HTTP errors
            ok_('X-Ratelimit-Limit' not in r.headers)
            ok_('X-Ratelimit-Remaining' not in r.headers)
            ok_('X-Ratelimit-Reset' not in r.headers)
            ok_('Retry-After' not in r.headers)

    def test_ratelimit(self):
        # Reset usage count for all users before beginning the test
        self.check('/ratelimit/reset')
        self.check('/ratelimit/reset?user=alpha')
        self.check('/ratelimit/reset?user=beta')

        # All HTTP headers come through OK. First usage leaves us with 2 remaining
        self.check_rate('/ratelimit/a', None, 3, 2)
        # /ratelimit/b uses the same pool as /a
        self.check_rate('/ratelimit/b', None, 3, 1)
        # HTTP errors don't deduct from usage
        self.check_rate('/ratelimit/a?n=0', None, 3, 1, INTERNAL_SERVER_ERROR)
        self.check_rate('/ratelimit/b?n=0', None, 3, 1, INTERNAL_SERVER_ERROR)
        # When usage exceeds, we get HTTP 429
        self.check_rate('/ratelimit/a', None, 3, 0)
        self.check_rate('/ratelimit/a', None, 3, 0, TOO_MANY_REQUESTS)
        self.check_rate('/ratelimit/b', None, 3, 0, TOO_MANY_REQUESTS)
        # Different users get a different limits
        self.check_rate('/ratelimit/a', 'alpha', 3, 2)
        self.check_rate('/ratelimit/a', 'beta', 3, 2)
        self.check_rate('/ratelimit/b', 'alpha', 3, 1)
        self.check_rate('/ratelimit/b', 'beta', 3, 1)

    def test_ratelimit_key_function(self):
        self.check('/ratelimit/reset3')
        self.check('/ratelimit/reset3?user=alpha')
        self.check('/ratelimit/reset3?user=beta')
        # alpha has a 4 daily limit. Others (including beta): 3 hourly
        self.check_rate('/ratelimit/key', 'alpha', 4, 3, expiry='daily')
        self.check_rate('/ratelimit/key', 'alpha', 4, 2, expiry='daily')
        self.check_rate('/ratelimit/key', 'beta', 3, 2, expiry='hourly')
        self.check_rate('/ratelimit/key', 'beta', 3, 1, expiry='hourly')
        self.check_rate('/ratelimit/key', None, 3, 2, expiry='hourly')
        self.check_rate('/ratelimit/key', None, 3, 1, expiry='hourly')

    def test_multiple_ratelimit(self):
        self.check('/ratelimit/reset')
        self.check('/ratelimit/reset?user=alpha')
        self.check('/ratelimit/reset?user=beta')
        self.check('/ratelimit/reset2')

        # Pool: None: 0/3, alpha: 0/3, beta: 0/3, all: 0/4
        self.check_rate('/ratelimit/ab', None, 3, 2)
        # Pool: None: 1/3, alpha: 0/3, beta: 0/3, all: 1/4
        self.check_rate('/ratelimit/ab', 'alpha', 3, 2)
        # Pool: None: 1/3, alpha: 1/3, beta: 0/3, all: 2/4
        self.check_rate('/ratelimit/ab', 'beta', 4, 1)
        # Pool: None: 1/3, alpha: 1/3, beta: 1/3, all: 3/4
        self.check_rate('/ratelimit/ab', None, 4, 0)
        # Pool: None: 2/3, alpha: 1/3, beta: 1/3, all: 4/4
        self.check_rate('/ratelimit/ab', 'alpha', 4, 0, TOO_MANY_REQUESTS)

        # We can reset just a single pool
        self.check('/ratelimit/reset2')
        # Pool: None: 2/3, alpha: 2/3, beta: 1/3, all: 0/4
        self.check_rate('/ratelimit/ab', 'beta', 3, 1)

    def test_get_ratelimit(self):
        self.check('/ratelimit/reset')

        def check(limit, usage):
            result = self.check('/ratelimit/info').json()
            eq_(result['limit'], limit)
            eq_(result['usage'], usage)
            eq_(result['remaining'], max(limit - usage - 1, 0))
            ok_(result['expiry'] < 86400)

        check(limit=3, usage=0)
        check(limit=3, usage=1)
        check(limit=3, usage=2)
        self.check('/ratelimit/info', code=TOO_MANY_REQUESTS)



'''Defines command line services to install, setup and run apps.'''

import contextlib
import io
import os
import sys
import time
import yaml
import stat
import shlex
import string
import shutil
import datetime
import requests
from shutil import which
from pathlib import Path
from collections import Counter

from orderedattrdict import AttrDict
from orderedattrdict.yamlutils import AttrDictYAMLLoader

# B404:import_subprocess only developers can access this, not users
from subprocess import Popen, check_output, CalledProcessError  # noqa S404
from textwrap import dedent
from tornado.template import Template
from zipfile import ZipFile
import gramex
import gramex.license
from gramex.config import ChainConfig, PathConfig, variables, app_log, slug


def install(args, kwargs):
    '''
    usage: gramex install <app> <url> [--target=DIR]

    "app" is any name you want to locally call the application.

    "url" can be a:

    - local ZIP file (/path/to/app.zip)
    - local directory (/path/to/directory)
    - URL of a ZIP file (https://github.com/user/repo/archive/master.zip)

    target is the directory to install at (defaults to user data directory.)

    After installation, runs "gramex setup" which runs the Makefile, setup.ps1,
    setup.sh, requirements.txt, setup.py, bower install, npm install, yarn install.

    Options:

    - `--cmd="COMMAND"` is a shell command to run. Any word "TARGET" in caps
      is substituted by the target directory.

    Installed apps:
    {apps}
    '''
    if len(args) < 1:
        app_log.error(show_usage('install'))
        return

    appname = args[0]
    app_log.info(f'Installing: {appname}')
    app_config = get_app_config(appname, kwargs)
    if len(args) == 2:
        app_config.url = args[1]
        run_install(app_config)
    elif 'url' in app_config:
        run_install(app_config)
    elif 'cmd' in app_config:
        returncode = run_command(app_config)
        if returncode != 0:
            app_log.error(f'Command failed with return code {returncode}. Aborting installation')
            return
    else:
        app_log.error(f'Use --url=... or --cmd=... to specific source of {appname}')
        return

    # Post-installation
    app_config.target = run_setup(app_config.target)
    app_config['installed'] = {'time': datetime.datetime.now(datetime.timezone.utc).isoformat()}
    save_user_config(appname, app_config)
    app_log.info(f'Installed. Run `gramex run {appname}`')


def setup(args, kwargs):
    '''
    usage: gramex setup <target> [<target> ...] [--all]

    target is the directory to set up (required). This can be an absolute path,
    relative path, or a directory name under $GRAMEXPATH/apps/.

    gramex setup --all sets up all apps under $GRAMEXPATH/apps/

    Run the following commands at that directory in sequence, if possible:

    - make
    - powershell -File setup.ps1
    - bash setup.sh (or sh setup.sh or ash setup.sh)
    - pip install --upgrade -r requirements.txt
    - python setup.py
    - bower --allow-root install
    - npm install
    - yarn install --prefer-offline
    '''
    for target in args:
        run_setup(target)
        return
    if 'all' in kwargs:
        root = os.path.join(variables['GRAMEXPATH'], 'apps')
        for filename in os.listdir(root):
            target = os.path.join(root, filename)
            # Only run setup on directories. Ignore __pycache__, etc
            if os.path.isdir(target) and not filename.startswith('_'):
                run_setup(target)
        return
    app_log.error(show_usage('setup'))


def uninstall(args, kwargs):
    '''
    usage: gramex uninstall <app> [<app> ...]

    "app" is the name of the locally installed application. You can uninstall
    multiple applications in one command.

    All information about the application is lost. You cannot undo this.

    Installed apps:
    {apps}
    '''
    if len(args) < 1:
        app_log.error(show_usage('uninstall'))
        return
    if len(args) > 1 and kwargs:
        app_log.errorf(f'Arguments allowed only with single app. Ignoring {", ".join(args[1:])}')
        args = args[:1]

    for appname in args:
        app_log.info(f'Uninstalling: {appname}')

        # Delete the target directory if it exists
        app_config = get_app_config(appname, kwargs)
        if os.path.exists(app_config.target):
            safe_rmtree(app_config.target)
        else:
            app_log.error(f'No directory {app_config.target} to remove')
        save_user_config(appname, None)


def run(args, kwargs):
    '''
    usage: gramex run <app> [--target=DIR] [--dir=DIR] [--<options>=<value>]

    "app" is the name of the locally installed application.

    If "app" is not installed, specify --target=DIR to run from DIR. The next
    "gramex run app" will automatically run from DIR.

    "dir" is a *sub-directory* under "target" to run from. This is useful if
    "app" has multiple sub-applications.

    All Gramex command line options can be used. These are saved. For example:

        gramex run app --target=/path/to/dir --listen.port=8899 --browser=true

    ... will preserve the "target", "listen.port" and "browser" values. Running
    "gramex run app" will re-use these values. To clear the option, leave the
    value blank. For example "--browser=" will clear the browser option.

    Installed apps:
    {apps}
    '''
    if len(args) < 1:
        app_log.error(show_usage('run'))
        return
    if len(args) > 1:
        app_log.error(f'Can only run one app. Ignoring {", ".join(args[1:])}')

    appname = args.pop(0)
    app_config = get_app_config(appname, kwargs)

    target = app_config.target
    if 'dir' in app_config:
        target = os.path.join(target, app_config.dir)
    if os.path.isdir(target):
        os.chdir(target)
        gramex.paths['base'] = Path('.')
        # If we run with updated parameters, save for next run under the .run config
        run_config = app_config.setdefault('run', {})
        for key in kwargs:
            if key not in app_keys:
                run_config[key] = app_config.pop(key)
        save_user_config(appname, app_config)
        # Tell the user what configs are used
        cline = ' '.join('--%s=%s' % arg for arg in flatten_config(app_config.get('run', {})))
        app_log.info(
            'Gramex %s | %s %s | %s | Python %s',
            gramex.__version__,
            appname,
            cline,
            os.getcwd(),
            sys.version.replace('\n', ' '),
        )
        gramex.init(args=AttrDict(app=app_config['run']))
    elif appname in apps_config['user']:
        # The user configuration has a wrong path. Inform user
        app_log.error(
            f'{appname}: no target path {app_config.target}. '
            f'Run "gramex uninstall {appname}" and try again.',
        )
    else:
        app_log.error(f'{appname}: no target path {app_config.target}')


def service(args, kwargs):
    '''
    usage: gramex service <cmd> [--options]

    Install a Gramex application as a Windows service

        gramex service install
            --cwd  "C:/path/to/application/"
            --user "DOMAIN\\USER"                # Optional user to run as
            --password "user-password"          # Required if user is specified
            --startup manual|auto|disabled

    Update the Windows service

        gramex service update <same parameters as install>

    Remove the Windows service

        gramex service remove       # or gramex service uninstall

    Start / stop commands

        gramex service start
        gramex service stop
    '''
    try:
        import gramex.winservice
    except ImportError:
        app_log.error('Unable to load winservice. Is this Windows?')
        raise
    if len(args) < 1:
        app_log.error(show_usage('service'))
        return
    gramex.winservice.GramexService.setup(args, **kwargs)


def init(args, kwargs):
    '''
    usage: gramex init [minimal] [--target=<path>]

    Initializes a Gramex project at the current directory or target path.
    Specifically, it:

    - Sets up a git repo
    - Install supporting files for a Gramex project from a template
      - "gramex init" sets up all dependencies
      - "gramex init minimal" sets up minimal dependencies
    - Runs gramex setup (which runs npm install and other dependencies)

    Options:

    - `--target <path>` is the location to install at. Defaults to $GRAMEXAPPS
    '''
    if len(args) > 2:
        app_log.error(show_usage('init'))
        return
    if len(args) == 0:
        args.append('default')
    source_dir = os.path.join(variables['GRAMEXPATH'], 'apps', 'init', args[0])
    if not os.path.exists(source_dir):
        app_log.error(f'Unknown init template: {args[0]} under {source_dir}')

    kwargs.setdefault('target', os.getcwd())
    app_log.info(f'Initializing Gramex project at {kwargs.target}')
    data = {
        'appname': os.path.basename(kwargs.target),
        'author': _check_output('git config user.name', default='Author'),
        'email': _check_output('git config user.email', default='user@example.org'),
        'date': datetime.datetime.today().strftime('%Y-%m-%d'),
        'version': gramex.__version__,
    }
    # Ensure that appname is a valid Python module name
    appname = slug.module(data['appname'])
    if appname[0] not in string.ascii_lowercase:
        appname = 'app' + appname
    data['appname'] = appname

    # Create a git repo. But if git fails, do not stop. Continue with the rest.
    with contextlib.suppress(OSError):
        _run_console('git init')
    # Install Git LFS if available. Set git_lfs=None if it fails, so .gitignore ignores assets/**
    data['git_lfs'] = which('git-lfs')
    if data['git_lfs']:
        try:
            _run_console('git lfs install')
            _run_console('git lfs track "assets/**"')
        except OSError:
            data['git_lfs'] = None

    # Copy all directories & files. Files with '.template.' are treated as templates.
    for root, dirs, files in os.walk(source_dir):
        relpath = os.path.relpath(root, start=source_dir)
        for name in dirs + files:
            source, targetname, template_data = os.path.join(root, name), name, None
            if '.template.' in name:
                targetname, template_data = name.replace('.template.', '.'), data
            target = os.path.join(kwargs.target, relpath, targetname)
            if os.path.exists(target):
                app_log.warning(f'Skip existing {target}')
            elif os.path.isdir(source):
                _mkdir(target)
            elif os.path.isfile(source):
                app_log.info(f'Copy file {source}')
                with io.open(source, 'rb') as handle:
                    result = handle.read()
                    if template_data is not None:
                        result = Template(result).generate(**template_data)
                with io.open(target, 'wb') as handle:
                    handle.write(result)
            else:
                app_log.warning(f'Skip unknown file {source}')

    run_setup(kwargs.target)


default_mail_config = r'''# Gramex mail configuration at
# List keys with "gramex mail --list --conf={confpath}"

# See https://gramener.com/gramex/guide/email/ for help
email:
  default-email:
    type: gmail
    email: $GRAMEXMAILUSER
    password: $GRAMEXMAILPASSWORD
    # Add stub: log to test the application without sending mails

# See https://gramener.com/gramex/guide/alert/
alert:
  hello-world:
    to: admin@example.org
    subject: Alert from Gramex
    body: |
      This is a test email
'''


def mail(args, kwargs):
    '''
    usage

        gramex mail <key>           # Send mail named <key>
        gramex mail --list          # Lists all keys in config file
        gramex mail --init          # Initializes config file

    The config is a gramex.yaml file. It must have email: and alert: sections.
    If the current folder has a gramex.yaml, that's used. Else the default is
    $GRAMEXDATA/mail/gramexmail.yaml.

    Options:

    - `--conf <path>` specifies a different conf file location
    '''
    # Get config file location
    default_dir = os.path.join(variables['GRAMEXDATA'], 'mail')
    _mkdir(default_dir)
    if 'conf' in kwargs:
        confpath = kwargs.conf
    elif os.path.exists('gramex.yaml'):
        confpath = os.path.abspath('gramex.yaml')
    else:
        confpath = os.path.join(default_dir, 'gramexmail.yaml')

    if not os.path.exists(confpath):
        if 'init' in kwargs:
            with io.open(confpath, 'w', encoding='utf-8') as handle:
                handle.write(default_mail_config.format(confpath=confpath))
            app_log.info(f'Initialized {confpath}')
        elif not args and not kwargs:
            app_log.error(show_usage('mail'))
        else:
            app_log.error(f'Missing config {confpath}. Use --init to generate skeleton')
        return

    conf = PathConfig(confpath)
    if 'list' in kwargs:
        for key, alert in conf.get('alert', {}).items():
            to = alert.get('to', '')
            if isinstance(to, list):
                to = ', '.join(to)
            gramex.console('{:15}\t"{}" to {}'.format(key, alert.get('subject'), to))
        return

    if 'init' in kwargs:
        app_log.error(f'Config already exists at {confpath}')
        return

    if len(args) < 1:
        app_log.error(show_usage('mail'))
        return

    from gramex.services import email as setup_email, create_alert

    alert_conf = conf.get('alert', {})
    email_conf = conf.get('email', {})
    setup_email(email_conf)
    sys.path += os.path.dirname(confpath)
    for key in args:
        if key not in alert_conf:
            app_log.error(f'Missing key {key} in {confpath}')
            continue
        alert = create_alert(key, alert_conf[key])
        alert()


def license(args, kwargs):
    '''
    usage: gramex license [accept|reject]

    Commands

        gramex license          # Show Gramex license
        gramex license accept   # Accept Gramex license
        gramex license reject   # Reject Gramex license
    '''
    if len(args) == 0:
        gramex.console(gramex.license.EULA)
        if gramex.license.is_accepted():
            gramex.console('License already ACCEPTED. Run "gramex license reject" to reject')
        else:
            gramex.console('License NOT YET accepted. Run "gramex license accept" to accept')
    elif args[0] == 'accept':
        gramex.license.accept(force=True)
    elif args[0] == 'reject':
        gramex.license.reject()
    else:
        app_log.error(f'Invalid command license {args[0]}')


def features(args, kwargs) -> dict:
    '''
    usage: gramex features [<path> [...]]

    Counts gramex features used in a project path (or current directory) from gramex.yaml.
    Prints 3 columns:

    - type: SVC=Service, MS=Microservice, KWARG=Config (keyword argument)
    - feature: Feature name (e.g. FormHandler)
    - count: Number of times the feature is used

    Options:

    - `--format [table|json|csv]` picks an output format. Defaults to table
    '''
    import pandas as pd

    # Load the feature mapping from cache
    features = gramex.cache.open("gramexfeatures.csv", rel=True).set_index(["type", "name"])[
        "feature"
    ]

    base = gramex.config.PathConfig(os.path.join(os.path.dirname(gramex.__file__), 'gramex.yaml'))
    # Load the gramex.yaml configuration
    usage = Counter({(type, feature): 0 for ((type, _name), feature) in features.items()})
    parsed_count = 0
    for folder in args or [os.getcwd()]:
        project_yaml = os.path.join(folder, 'gramex.yaml')
        app = gramex.config.PathConfig(project_yaml)
        # Skip if gramex.yaml is not found or empty or we're unable to parse
        if not app:
            continue
        parsed_count += 1
        conf = +gramex.config.ChainConfig([('base', base), ('app', app)])

        # Convert invalid gramex.yaml (e.g. empty gramex.yaml) as an empty dict
        conf = conf if isinstance(conf, dict) else {}
        # List the features used. TODO: Improve this using configuration
        for url in conf.get('url', {}).values():
            if isinstance(url, dict):
                name = url.get('handler', '').replace('gramex.handlers.', '')
                # Count micro-services usage
                if name in features['MS']:
                    usage[('MS', features['MS'][name])] += 1
                else:
                    usage[('ERR-MS', name)] += 1
                # Count kwargs: usage
                url_kwargs = url.get('kwargs', {})
                for key in url_kwargs:
                    if key in features['KWARG']:
                        usage[('KWARG', features['KWARG'][key])] += 1
                # cache: is a special kwarg that sits outside kwargs. Count it too
                if 'cache' in url:
                    usage[('KWARG', 'cache')] += 1
        # List the services used
        for service in conf.keys():
            # If the service is conditional, e.g. `log if condition`, take just the first part
            name = service.split()[0]
            if name in features['SVC']:
                usage[('SVC', features['SVC'][name])] += 1

    # Return features usage table
    output = ''
    if parsed_count:
        usage = pd.DataFrame(usage.values(), index=usage.keys()).reset_index()
        usage.columns = ['type', 'feature', 'count']
        format = kwargs.get('format', 'table')
        if format == 'table':
            output = usage.to_string()
        elif format == 'csv':
            output = usage.to_csv(index=False, lineterminator='\n')
        elif format == 'json':
            output = usage.to_json(orient='records')
        else:
            app_log.error(f'--format must be [table|csv|json] not {format}')
            return
    return output


def complexity(args, kwargs) -> dict:
    '''
    usage: gramex complexity [folder [...]]
    '''
    # Calulate cyclomatic complexity of the project
    import ast
    import fnmatch
    import mccabe
    import pandas as pd
    import re

    # B404:import_subprocess only used for internal Gramex scripts
    from subprocess import check_output  # noqa S404

    project_path = os.getcwd() if len(args) == 0 else args[0]
    project_yaml = _gramex_yaml_path(project_path, kwargs)
    if project_yaml is None:
        return

    def walk(node: dict, parents: tuple = ()):
        for key, value in node.items():
            new_key = parents + (str(key),)
            if hasattr(value, 'items'):
                yield new_key, None
                yield from walk(value, new_key)
            else:
                yield new_key, value

    py_complexity = 0
    relevent_file_matcher = re.compile(r'^((?!(site-packages|node_modules)).)*\.py$')

    for root, _dirs, files in os.walk(project_path):
        for filename in files:
            path: str = os.path.join(root, filename)
            if not relevent_file_matcher.match(path):
                continue
            rel_path: str = os.path.relpath(path, project_path)
            with open(path, 'rb') as handle:
                code = handle.read()
            try:
                tree = compile(code, rel_path, 'exec', ast.PyCF_ONLY_AST)
            except SyntaxError as e:
                app_log.exception('SYNTAXERROR: %s %s', path, e)
                continue
            visitor = mccabe.PathGraphingAstVisitor()
            visitor.preorder(tree, visitor)
            for node in visitor.graphs.values():
                py_complexity += node.complexity()

    base = gramex.config.PathConfig(os.path.join(os.path.dirname(gramex.__file__), 'gramex.yaml'))
    try:
        app = gramex.config.PathConfig(project_yaml)
        conf = +gramex.config.ChainConfig([('base', base), ('app', app)])
    except Exception as e:  # noqa B902 capture load errors as a "feature"
        app_log.exception(str(e))
        return
    yamlpaths = {'.'.join(key): val for key, val in walk(conf)}
    used = set()
    gramexsize = gramex.cache.open('gramexsize.csv', rel=True)
    for _index, gramex_code in gramexsize.iterrows():
        # TODO: fnmatch.filter() is the slowest part. Optimize it
        if '=' not in gramex_code['yamlpath']:
            if fnmatch.filter(yamlpaths.keys(), gramex_code['yamlpath']):
                used.add(gramex_code['codepath'])
        else:
            yaml_key, yaml_value = gramex_code['yamlpath'].split('=', 1)
            matches = fnmatch.filter(yamlpaths.keys(), yaml_key)
            for key in matches:
                if yamlpaths[key] in yaml_value.split('|'):
                    used.add(gramex_code['codepath'])
    gramex_complexity = gramexsize.set_index('codepath')['complexity'][list(used)].sum()

    # Calculate JS complexity
    # B602:subprocess_popen_with_shell_equals_true and
    # B607:start_process_with_partial_path are safe to skip since this is a Gramex internal cmd
    output = check_output(  # noqa S602
        "npx --yes @gramex/escomplexity", cwd=project_path, shell=True  # noqa S607
    )
    es_complexity = int(output.decode('utf-8').split('\n')[-2].strip())
    return pd.DataFrame(
        {
            'py': [py_complexity],
            'js': [es_complexity],
            'gramex': [gramex_complexity],
        }
    )


class TryAgainError(Exception):
    '''If shutil.rmtree fails, and we've fixed the problem, raise this to try again'''

    pass


try:
    WindowsError
except NameError:
    # On non-Windows systems, _ensure_remove just raises the exception
    def _ensure_remove(remove, path, exc_info):
        raise exc_info[1]

else:
    # On Windows systems, try harder
    def _ensure_remove(func, path, exc_info):
        '''onerror callback for rmtree that tries hard to delete files'''
        if issubclass(exc_info[0], WindowsError):
            winerror = AttrDict(
                ERROR_PATH_NOT_FOUND=3,
                ERROR_ACCESS_DENIED=5,
                ERROR_SHARING_VIOLATION=32,
            )
            # Delete read-only files
            # https://bugs.python.org/issue19643
            # https://bugs.python.org/msg218021
            if exc_info[1].winerror == winerror.ERROR_ACCESS_DENIED:
                os.chmod(path, stat.S_IWRITE)
                return os.remove(path)
            # Delay delete a bit if directory is used by another process.
            # Typically happens on uninstall immediately after bower / npm / git
            # (e.g. during testing.)
            elif exc_info[1].winerror == winerror.ERROR_SHARING_VIOLATION:
                delays = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1.0]
                for delay in delays:
                    time.sleep(delay)
                    with contextlib.suppress(WindowsError):
                        return os.remove(path)
            # npm creates windows shortcuts that shutil.rmtree cannot delete.
            # os.listdir/scandir fails with a PATH_NOT_FOUND.
            # Delete these using win32com and try again.
            elif exc_info[1].winerror == winerror.ERROR_PATH_NOT_FOUND and func in {
                os.listdir,
                os.scandir,
            }:
                app_log.error(f'Cannot delete {path}')
                from win32com.shell import shell, shellcon  # type:ignore

                options = shellcon.FOF_NOCONFIRMATION | shellcon.FOF_NOERRORUI
                code, err = shell.SHFileOperation((0, shellcon.FO_DELETE, path, None, options))
                if code == 0:
                    raise TryAgainError()
        raise exc_info[1]


def _try_remove(target, retries=100, delay=0.05, func=shutil.rmtree, **kwargs):
    for _index in range(retries):
        try:
            func(target, **kwargs)
        except TryAgainError:
            pass
        # If permission is denied, e.g. antivirus, file is open, etc, keep trying with delay
        except OSError:
            app_log.warning('    Trying again to delete', target)
            time.sleep(delay)
        else:
            break


def safe_rmtree(target, retries=100, delay=0.05, gramexdata=True):
    '''
    A replacement for shutil.rmtree and os.remove that removes directories,
    optionally within $GRAMEXDATA.
    It tries to remove the target multiple times, recovering from errors.
    '''
    if not os.path.exists(target):
        return True
    # TODO: check case insensitive in Windows, but case sensitive on other OS
    func, kwargs = (
        (shutil.rmtree, {'onerror': _ensure_remove}) if os.path.isdir(target) else (os.remove, {})
    )
    if gramexdata:
        if target.lower().startswith(variables['GRAMEXDATA'].lower()):
            # Try multiple times to recover from errors, since we have no way of
            # auto-resuming rmtree: https://bugs.python.org/issue8523
            _try_remove(target, retries, delay, func, **kwargs)
            return True
        else:
            app_log.warning(f'Not removing directory {target} (outside $GRAMEXDATA)')
            return False
    else:
        _try_remove(target, retries, delay, func, **kwargs)


def zip_prefix_filter(members, prefix):
    '''Return only ZIP file members starting with directory prefix, with prefix stripped out.'''
    if not prefix:
        return members
    offset = len(prefix)
    result = []
    for zipinfo in members:
        if zipinfo.filename.startswith(prefix):
            zipinfo.filename = zipinfo.filename[offset:]
            if len(zipinfo.filename) > 0:
                result.append(zipinfo)
    return result


def run_install(config):
    '''
    Download config.url into config.target.
    If config.url is a directory, copy it.
    If config.url is a file or a URL (http, https, ftp), unzip it.
    If config.contentdir is True, skip parent folders with single subfolder.
    If no files match, log a warning.
    '''
    url, target = config.url, config.target

    # If the URL is a directory, copy it
    if os.path.isdir(url):
        if os.path.exists(target):
            url = os.path.abspath(url).lower().rstrip(os.sep)
            target = os.path.abspath(target).lower().rstrip(os.sep)
            if url != target and not safe_rmtree(target):
                return
        if url != target:
            shutil.copytree(url, target)
            app_log.info(f'Copied {url} into {target}')
        config.url = url
        return

    # If it's a file, unzip it
    if os.path.exists(url):
        handle = url
    else:
        # Otherwise, assume that it's a URL containing a ZIP file
        app_log.info(f'Downloading: {url}')
        response = requests.get(url, timeout=30)
        response.raise_for_status()
        handle = io.BytesIO(response.content)

    # Identify relevant files from the ZIP file
    zipfile = ZipFile(handle)
    files = zipfile.infolist()
    if config.get('contentdir', True):
        prefix = os.path.commonprefix(zipfile.namelist())
        if prefix.endswith('/'):
            files = zip_prefix_filter(files, prefix)

    # Extract relevant files from ZIP file
    if safe_rmtree(target):
        zipfile.extractall(target, files)
        app_log.info(f'Extracted {len(files)} files into {target}')


def run_command(config):
    '''
    Run config.cmd. If the command has a TARGET, replace it with config.target.
    Else append config.target as an argument.
    '''
    appcmd = config.cmd
    # Split the command into an array of words
    if isinstance(appcmd, str):
        appcmd = shlex.split(appcmd)
    # If the app is a Cygwin app, TARGET should be a Cygwin path too.
    target = config.target
    cygwin, cygpath, kwargs = which('cygcheck'), which('cygpath'), {'universal_newlines': True}
    if cygwin is not None and cygpath is not None:
        # subprocess.check_output is safe here since these are developer-initiated
        path = check_output([cygpath, '-au', which(appcmd[0])], **kwargs).strip()  # noqa S603
        is_cygwin_app = check_output([cygwin, '-f', path], **kwargs).strip()  # noqa S603
        if is_cygwin_app:
            target = check_output([cygpath, '-au', target], **kwargs).strip()  # noqa S603
    # Replace TARGET with the actual target
    if 'TARGET' in appcmd:
        appcmd = [target if arg == 'TARGET' else arg for arg in appcmd]
    else:
        appcmd.append(target)
    app_log.info(f'Running {" ".join(appcmd)}')
    if not safe_rmtree(config.target):
        app_log.error(f'Cannot delete target {config.target}. Aborting installation')
        return
    # B603:subprocess_without_shell_equals_true is safe since this is developer-initiated
    proc = Popen(appcmd, bufsize=-1, **kwargs)  # noqa S603
    proc.communicate()
    return proc.returncode


# Setup file configurations. If {file} exists, then if {exe} exists, run {cmd}.
setup_paths = [
    [{'file': 'Makefile', 'exe': 'make', 'cmd': '"{exe}"'}],
    [{'file': 'setup.ps1', 'exe': 'powershell', 'cmd': '"{exe}" -File "{file}"'}],
    [{'file': 'setup.sh', 'exe': ['bash', 'sh', 'ash'], 'cmd': '"{exe}" "{file}"'}],
    [{'file': 'requirements.txt', 'exe': 'pip', 'cmd': '"{exe}" install -r "{file}"'}],
    [{'file': 'setup.py', 'exe': 'python', 'cmd': '"{exe}" "{file}"'}],
    [{'file': 'bower.json', 'exe': 'bower', 'cmd': '"{exe}" --allow-root install'}],
    [
        {'file': 'package-lock.json', 'exe': 'npm', 'cmd': '"{exe}" ci'},
        {'file': 'yarn.lock', 'exe': 'yarn', 'cmd': '"{exe}" install --prefer-offline'},
        {'file': 'package.json', 'exe': 'npm', 'cmd': '"{exe}" install'},
        {'file': 'package.json', 'exe': 'yarn', 'cmd': '"{exe}" install --prefer-offline'},
    ],
]


def run_setup(target):
    '''
    Install any setup file in target directory. Target directory can be:

    - An absolute path
    - A relative path to current directory
    - A relative path to the Gramex apps/ folder

    Returns the absolute path of the final target path.
    '''
    if not os.path.exists(target):
        app_target = os.path.join(variables['GRAMEXPATH'], 'apps', target)
        if not os.path.exists(app_target):
            raise OSError(f'No directory {target}')
        target = app_target
    target = os.path.abspath(target)
    app_log.info(f'Setting up {target}')
    for configs in setup_paths:
        config_match, ran_cmd = None, False
        for config in configs:
            setup_file = os.path.join(target, config['file'])
            exes = config['exe'] if isinstance(config['exe'], list) else [config['exe']]
            exe_path = next((which(exe) for exe in exes if which(exe)), None)
            if os.path.exists(setup_file):
                config_match = config
            if config_match and exe_path is not None:
                cmd = config['cmd'].format(file=setup_file, exe=exe_path)
                app_log.info(f'Running {cmd}')
                _run_console(cmd, cwd=target)
                ran_cmd = True
                break
        if config_match and not ran_cmd:
            app_log.warning(f'Skipping {config_match["file"]}. No {config_match["exe"]} found')


app_dir = Path(variables.get('GRAMEXDATA')) / 'apps'
if not app_dir.exists():
    app_dir.mkdir(parents=True)

# Get app configuration by chaining apps.yaml in gramex + app_dir + command line
apps_config = ChainConfig()
apps_config['base'] = PathConfig(gramex.paths['source'] / 'apps.yaml')
user_conf_file = app_dir / 'apps.yaml'
apps_config['user'] = PathConfig(user_conf_file) if user_conf_file.exists() else AttrDict()

app_keys = {
    'url': 'URL / filename of a ZIP file to install',
    'cmd': 'Command used to install file',
    'dir': 'Sub-directory under "url" to run from (optional)',
    'contentdir': 'Strip root directory with a single child (optional, default=True)',
    'target': 'Local directory where the app is installed',
    'installed': 'Additional installation information about the app',
    'run': 'Runtime keyword arguments for the app',
}


def save_user_config(appname, value):
    user_config = AttrDict()
    if user_conf_file.exists():
        with user_conf_file.open(encoding='utf-8') as handle:
            # If app.yaml is empty, yaml.safe_load returns None. Use the AttrDict() instead
            # B506:yaml_load is safe since this object is internally created
            user_config = yaml.load(handle, Loader=AttrDictYAMLLoader) or user_config  # noqa S506
    if value is None:
        if appname in user_config:
            del user_config[appname]
    else:
        app_config = user_config.setdefault(appname, AttrDict())
        app_config.update({key: value[key] for key in app_keys if key in value})

    with user_conf_file.open(mode='w', encoding='utf-8') as handle:
        # Use yaml.dump (not .safe_dump) to serialize the AttrDicts
        # B506:yaml_load is safe since this object is internally created
        yaml.dump(user_config, handle, indent=4, default_flow_style=False)  # noqa S506


def get_app_config(appname, kwargs):
    '''
    Get the stored configuration for appname, and override it with kwargs.
    `.target` defaults to $GRAMEXDATA/apps/<appname>.
    '''
    apps_config['cmd'] = {appname: kwargs}
    app_config = AttrDict((+apps_config).get(appname, {}))
    app_config.setdefault('target', str(app_dir / app_config.get('target', appname)))
    app_config.target = os.path.abspath(app_config.target)
    return app_config


def flatten_config(config, base=None):
    'Get flattened configurations'
    for key, value in config.items():
        keystr = key if base is None else base + '.' + key
        if hasattr(value, 'items'):
            yield from flatten_config(value, keystr)
        else:
            yield keystr, value


def show_usage(command):
    apps = (+apps_config).keys()
    return usage[command].format(apps='\n'.join('- ' + app for app in sorted(apps)))


def _check_output(cmd, default=b'', **kwargs):
    '''Run cmd and return output. Return default in case the command fails'''
    try:
        # B603:subprocess_without_shell_equals_true is safe since this is developer-initiated
        return check_output(shlex.split(cmd), **kwargs).strip()  # noqa S603
    # OSError is raised if the cmd is not found.
    # CalledProcessError is raised if the cmd returns an error.
    except (OSError, CalledProcessError):
        return default


def _run_console(cmd, **kwargs):
    '''Run cmd and pipe output to console. Log and raise error if cmd is not found'''
    cmd = shlex.split(cmd)
    try:
        # B603:subprocess_without_shell_equals_true is safe since this is developer-initiated
        proc = Popen(cmd, bufsize=-1, universal_newlines=True, **kwargs)  # noqa S603
    except OSError:
        app_log.error(f'Cannot find command: {cmd[0]}')
        raise
    proc.communicate()


def _mkdir(path):
    '''Create directory tree up to path if path does not exist'''
    if not os.path.exists(path):
        os.makedirs(path)


def _gramex_yaml_path(folder, kwargs):
    project_yaml = os.path.join(folder, 'gramex.yaml')
    # Get config file location
    if 'conf' in kwargs:
        return kwargs.conf
    elif os.path.exists(project_yaml):
        return project_yaml
    else:
        app_log.error(f'Missing {project_yaml}')
        return


usage = {
    'complexity': dedent(complexity.__doc__),
    'features': dedent(features.__doc__),
    'init': dedent(init.__doc__),
    'install': dedent(install.__doc__),
    'license': dedent(license.__doc__),
    'mail': dedent(mail.__doc__),
    'run': dedent(run.__doc__),
    'service': dedent(service.__doc__),
    'setup': dedent(setup.__doc__),
    'uninstall': dedent(uninstall.__doc__),
}
